#!/usr/bin/env python

import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

matplotlib.rcParams['font.size'] = 10 #'small'
matplotlib.rcParams['font.sans-serif'] = 'Arial'
matplotlib.rcParams['axes.labelsize'] = 10 #'small'
matplotlib.rcParams['axes.titlesize'] = 10 
matplotlib.rcParams['xtick.labelsize'] = 8 #'small'
matplotlib.rcParams['ytick.labelsize'] = 8 #'small'
matplotlib.rcParams['legend.fontsize'] = 9

import numpy
import os
import sys
import glob

###
sample_rename = {
    'PTB_CLIP': 'CLIP PTB', 
    'PTB_rep3': 'iCLIP PTB',

    'SRSF3_rep2': 'SRSF3',
    'SRSF4_rep1': 'SRSF4',

    'TIAL1_rep4': 'TIAL1',

    'U2AF65': 'U2AF65',

    'eIF4A3_CLIP': 'CLIP eIF4A3',
    'eIF4A3_rep5': 'iCLIP eIF4A3',

    'hnRNPL': 'hnRNP L',
}
###

# read graph information
def read_one_set(fin_name):
    retd = {}
    flank = 60
    xs = [x for x in range(-flank, 0) + range(flank+1)]
    expected_header = [ "reads_group", "reference_group", "origin"]
    expected_header = expected_header + [str(x) for x in xs]
    fin = open(fin_name, "rt")
    r = fin.readline()
    assert r.rstrip('\n\r').split('\t') == expected_header
    r = fin.readline()
    while r:
        rs = r.rstrip('\n\r').split('\t')
        assert len(rs) == len(expected_header)
        read_grp, ref_grp, origin = rs[:3]
        freqs = [int(x) for x in rs[3:]]
        retd[(read_grp, ref_grp, origin)] = (xs, freqs)
        r = fin.readline()
    fin.close() 
    return retd

data = {}
for fn in glob.glob("results/*_graphs.tab"):
    prot_id = fn.split('/')[-1][:-len('_graphs.tab')]
    if prot_id not in sample_rename:
        print "skipping:", prot_id
        continue
    prot_id = sample_rename[prot_id]
    assert prot_id not in data
    data[prot_id] = read_one_set(fn)

print "read info on %s proteins: %s" % (len(data), sorted(data.keys()))

## generate selected figures
def matrix_plot(fout_name, row_names, col_names_top, col_names_bottom, m, left=0.15, right=0.9, bottom=0.08, top=0.9, wspace=0.55, hspace=0.7):
    print "generating:", fout_name
    rows = len(row_names)
    cols = len(col_names_top)
    assert cols == len(col_names_bottom)
    assert len(m) == rows
    rows += 1

    plt.figure(figsize=(cols*1.9, rows*1.7))
    plt.clf()
    plt.subplots_adjust(left=left, right=right, bottom=bottom, top=top, wspace=wspace, hspace=hspace)

    si = 1
    max_pad = 0.0
    for row_i, (row_name, r) in enumerate(zip(row_names, m)):
        assert len(r) == cols
        for col_i, (col_name_top, col_name_bottom, (xs, ys)) in enumerate(zip(col_names_top, col_names_bottom, r)):
            ax = plt.subplot(rows, cols, si)
            si += 1
            # RGB: 0, 129, 197 (0081c5)
            # HSB: 201, 100, 77
            plt.bar(xs, ys, width=0.7, ec='#0081c5', color='#0081c5')
            plt.xlim((xs[0], xs[-1]))

            ylim = plt.ylim()    
            plt.plot([0, 0], ylim, c='k', alpha=0.3) #lw=0.8)
            plt.ylim(ylim)
            yticks_vals, yticks_labels = plt.yticks()
            yticks_max = yticks_vals[-1]
            yticks = [0, yticks_max*1.0/4.0, yticks_max*2.0/4.0, yticks_max*3.0/4.0, yticks_max]
            #yticks = [0, yticks_max*1.0/3.0, yticks_max*2.0/3.0, yticks_max]
            plt.yticks(yticks, ["%0.0f" % y for y in yticks])
            plt.xticks([-60, -30, 0, 30, 60], ["-60", "-30", "0", "30", "60"])

            if row_i == 0:
                plt.title(col_name_top)
            if col_i == 0:
                plt.ylabel(row_name, multialignment='center', verticalalignment='center') #, horizontalalignment='right')
                ax.get_yaxis().set_label_coords(-0.4,0.5)

            if row_i == len(row_names) - 1:
                plt.xlabel(col_name_bottom, multialignment='center') #, verticalalignment='baseline', horizontalalignment='right')
            
    plt.savefig(fout_name, bbox_inches='tight')
    plt.clf()

# figure 4: all iCLIP libaries (excluding BTZ): A1, A2, A3 reference to B
row_names = []
prot_order = ['TIAL1', 'U2AF65', 'iCLIP eIF4A3', 'SRSF3', 'SRSF4', 'iCLIP PTB', 'hnRNP L']

m = []
for prot_id in prot_order:
    tmpd = data[prot_id]
    r = []
    for t in ["A1", "A2", "A3"]:
        k = (t, "B", "start")
        r.append(tmpd[k])
    m.append(r)
    row_names.append(prot_id)

matrix_plot("results/fig4.pdf", row_names,
            ["Offset of A1 to B\n start sites\n",
            "Offset of A2 to B\n start sites\n",
            "Offset of A3 to B\n start sites\n",
            ],
            ["\nOffset of A1 to B\n start sites",
            "\nOffset of A2 to B\n start sites",
            "\nOffset of A3 to B\n start sites",
            ], m
            )

# figure 5: eIF4A3 CLIP vs iCLIP: A1, A2 reference to A3 (total only, excluding exon and intron)
row_names = []
prot_order = ['iCLIP eIF4A3', 'CLIP eIF4A3']
m = []
for orig in ["start", "center"]:
    r = []
    for prot_id in prot_order:
        tmpd = data[prot_id]
        for t in ["A1", "A2"]:
            k = (t, "A3", orig)
            r.append(tmpd[k])
    m.append(r)
    row_names.append(orig)

matrix_plot("results/fig5.pdf", row_names,
            ["iCLIP eIF4A3\n", "", "CLIP eIF4A3\n", ""],
            ["Offset of A1 to A3",
            "\nOffset of A2 to A3",
            "\nOffset of A1 to A3",
            "\nOffset of A2 to A3",
            ], m
            )

# figure S3: PTB CLIP vs iCLIP: A1, A2 reference to A3 (total only, excluding exon and intron)
row_names = []
prot_order = ['iCLIP PTB', 'CLIP PTB']
m = []
for orig in ["start", "center"]:
    r = []
    for prot_id in prot_order:
        tmpd = data[prot_id]
        for t in ["A1", "A2"]:
            k = (t, "A3", orig)
            r.append(tmpd[k])
    m.append(r)
    row_names.append(orig)

matrix_plot("results/figS3.pdf", row_names,
            ["iCLIP PTB\n", "", "CLIP PTB\n", ""],
            ["\nOffset of A1 to\nA3 sites",
            "\nOffset of A2 to\nA3 sites",
            "\nOffset of A1 to\nA3 sites",
            "\nOffset of A2 to\nA3 sites",
            ], m
            )

# figure S2: A1, A2 reference A3 and A1, A2, A3 reference B, for start and center, for different iCLIP experiments: (a) SRSF3, (b) SRSF4, (c) hnRNP L (d) TIAL1 and (e) U2AF65.
row_names = []
prot_order = ['SRSF3', 'SRSF4', 'hnRNP L', 'TIAL1', 'U2AF65']
m = []
for prot_id in prot_order:
    tmpd = data[prot_id]
    for orig in ["start", "center"]:
        r = []
        for t in ["A1", "A2"]:
            k = (t, "A3", orig)
            r.append(tmpd[k])
        for t in ["A1", "A2", "A3"]:
            k = (t, "B", orig)
            r.append(tmpd[k])
        m.append(r)
        row_names.append("%s\n%s" % (prot_id, orig))

matrix_plot("results/figS2.pdf", row_names,
            ["Offset of A1 to A3\n",
            "Offset of A2 to A3\n",
            "Offset of A1 to B\n",
            "Offset of A2 to B\n",
            "Offset of A3 to B\n",
            ], 

            ["\nOffset of A1 to A3",
            "\nOffset of A2 to A3",
            "\nOffset of A1 to B",
            "\nOffset of A2 to B",
            "\nOffset of A3 to B",
            ], m
            )

print "DONE"

