"""iCLIPro is a package that tests iCLIP read overlaps.
It can aid in the indentification and correction of
potential systematic errors in iCLIP data.

For documentation, see http://www.biolab.si/iCLIPro

Documentation source can be found in folder docs.

Examples are provided in folder examples.
"""

from _version import __version__

import iCLIPro.bed

# change this to True if you want to see an experimental
# test where we count the number of sites identified in
# each interpretation (start, center)
experimental = False

import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

matplotlib.rcParams['font.size'] = 10 #'small'
matplotlib.rcParams['font.sans-serif'] = 'Arial'
matplotlib.rcParams['axes.labelsize'] = 10 #'small'
matplotlib.rcParams['axes.titlesize'] = 10
matplotlib.rcParams['xtick.labelsize'] = 8 #'small'
matplotlib.rcParams['ytick.labelsize'] = 8 #'small'
matplotlib.rcParams['legend.fontsize'] = 9

import os
import random
import pysam

def amb_match(s1, s2):
    #assert len(s1) == len(s2)
    cn = sum([(c1 == 'N' or c2 == 'N' or c1 == c2) for c1, c2 in zip(s1, s2)])
    return len(s1) - cn <= 2

def collapse_bcs(reads):
    retl = []
    # group reads by starting (xlink_pos)
    by_pos = {}
    for xlink_pos, middle_pos, bc in reads:
        by_pos.setdefault(xlink_pos, {}).setdefault(bc, []).append(middle_pos)

    # reassign reads with ambiguous rnd barcodes to most frequent unambiguous matching rnd barcode
    # determine all unambiguous rnd barcodes
    for xlink_pos, by_bc in by_pos.iteritems():
        accepted_bcs = []
        ambig_bcs = []
        for bc, hits in by_bc.iteritems():
            Ns = bc.count("N")
            if Ns == 0:
                accepted_bcs.append((len(hits), bc))
            else:
                ambig_bcs.append((Ns, bc))
        # sort unambiguous by decreasing frequency (number of reads it is present)
        accepted_bcs = [bc for _, bc in sorted(accepted_bcs, reverse=True)]

        # add those ambiguous (in increasing order of the number of Ns in rnd barcode)
        ambig_bcs.sort()
        for _, abc in ambig_bcs:
            matches = False
            for bc in accepted_bcs:
                if amb_match(abc, bc):
                    matches = True
                    # move read records from abc under bc
                    by_bc[bc].extend(by_bc[abc])
                    del by_bc[abc]
                    break
            if not matches:
                accepted_bcs.append(abc)

        for bc, middle_poss in by_bc.iteritems():
            middle_pos = random.choice(middle_poss)
            retl.append((xlink_pos, middle_pos))

    return retl

def overlap_test(in_bam_fname, output_folder, bin_size, bin_min_reads, neighbor_clustering,
                 readlen2groups, group_comparisons,
                 mapq_th=30, split_flank=5, group_comparisons_flank=60):

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    bam_basename = os.path.basename(in_bam_fname).rsplit('.', 1)[0]
    out_fname_pref = "%s/%s" % (output_folder, bam_basename)

    ### read file
    try:
        samfile = pysam.Samfile(in_bam_fname, "rb")
    except:
        print "Error: could not open BAM file: %s" % in_bam_fname
        return None
    # can we rely on direct indexing to get chrome name
    assert [samfile.getrname(i) == rname for i, rname in enumerate(samfile.references)]

    # statistics counters
    all_recs = 0
    notmapped_recs = 0
    mapped_recs = 0
    lowmapq_recs = 0 
    # ignore reads with gaps at beginning or end of reads,
    # because hard to tell if exon-intron or exon-exon
    split_flank_ignored_recs = 0
    used_recs = 0

    # assign reads to bins
    valid_bc_nucs = set('ATCGN')
    bins = {}
    all_bcs = {} # list of all random barcodes and used to reduce memory use
    split_flank_distance = split_flank - 1
    for r in samfile:
        all_recs += 1
        #if all_recs > 500000: break #DEBUG
        if r.is_unmapped:
            notmapped_recs += 1
            continue

        if r.mapq < mapq_th:
            lowmapq_recs += 1
            continue

        mapped_recs += 1

        ## identify gaps
        # list of positions to which read mapped to
        poss = sorted(r.positions)

        # ignore reads that have gaps whithin the first/last (split_flank) nucleotides
        # longest gap between consecutive nucleotides
        gaps = [y-x for x, y in zip(poss, poss[1:])]

        # two consecutive nucleotides are 1 apart
        if max(gaps[:split_flank_distance]+[0] + gaps[-split_flank_distance:]) > 1:
            split_flank_ignored_recs += 1
            continue

        ## use this read record because it passes all filtering steps
        used_recs += 1

        ## random barcode must be part of read name
        if ':rbc' in r.qname:
            bc = r.qname.rsplit(':rbc', 1)[1].split(':',)[0]
        elif ':' in r.qname:
            bc = r.qname.rsplit(':', 1)[1]
            if set(bc) - valid_bc_nucs:
                # barcode contains invalid characters
                bc = ''
        else:
            bc = ''
        bc = all_bcs.setdefault(bc, bc)

        ## position of cross-link is one nucleotide before start of read
        if r.is_reverse:
            strand = '-'
            xlink_pos = poss[-1] + 1
        else:
            strand = '+'
            xlink_pos = poss[0] - 1
        chrome = samfile.references[r.tid]

        ## middle position is position of middle nucleotide
        # (should not take just middle of mapped region, because we can have spliced reads)
        middle_pos = poss[len(poss)/2]
        read_len = len(r.seq)

        ## place reads in bins
        bin_index = xlink_pos / bin_size
        bins.setdefault((chrome, strand), {}).setdefault(bin_index, []).append((xlink_pos, middle_pos, bc, read_len))
    samfile.close()

    ### identify bins with enough reads
    tot_bins = 0
    tot_reads = 0
    selected_reads = 0
    selected_reads_len_distrib = {}
    selected_bins = 0
    bins_to_remove = []
    for (chrome, strand), by_bin_index in bins.iteritems():
        tot_bins += len(by_bin_index)

        for bin_index, reads in by_bin_index.iteritems():
            reads_in_bin = len(reads)

            tot_reads += reads_in_bin
            if reads_in_bin >= bin_min_reads:
                selected_bins += 1
                selected_reads += reads_in_bin
                for _, _, _, read_len in reads:
                    selected_reads_len_distrib[read_len] = selected_reads_len_distrib.get(read_len, 0) + 1
            else:
                bins_to_remove.append((chrome, strand, bin_index))
    max_read_len = max(selected_reads_len_distrib.keys())
    # remove bins that are not needed anymore
    for chrome, strand, bin_index in bins_to_remove:
        tmpd = bins[(chrome, strand)]
        del tmpd[bin_index]

    ### generate bedGraphs for each group
    g2lens = {}
    for read_len, read_grps in readlen2groups.iteritems():
        for read_grp in read_grps:
            g2lens.setdefault(read_grp, []).append(read_len)

    grp_bed_start = {}
    grp_bed_middle = {}
    for read_grp, valid_read_lens in g2lens.iteritems():
        bed_start = grp_bed_start.setdefault(read_grp, {})
        bed_middle = grp_bed_middle.setdefault(read_grp, {})

        for (chrome, strand), by_bin_index in bins.iteritems():
            tmpd_bed_start = bed_start.setdefault((chrome, strand), {})
            tmpd_bed_middle = bed_middle.setdefault((chrome, strand), {})

            for bin_index, reads in by_bin_index.iteritems():
                valid_reads = [(xlink_pos, middle_pos, bc) for xlink_pos, middle_pos, bc, read_len in reads if read_len in valid_read_lens]

                for xlink_pos, middle_pos in collapse_bcs(valid_reads):
                    tmpd_bed_start[xlink_pos] = tmpd_bed_start.get(xlink_pos, 0) + 1
                    tmpd_bed_middle[middle_pos] = tmpd_bed_middle.get(middle_pos, 0) + 1
    # bins not needed after this point
    del bins

    ### save bedGraphs
    for read_grp, bed_start in sorted(grp_bed_start.iteritems()):
        iCLIPro.bed.save("%s_%s_start.bed" % (out_fname_pref, read_grp), bed_start, "%s start %s" % (read_grp, bam_basename), db_name="")
    for read_grp, bed_middle in sorted(grp_bed_middle.iteritems()):
        iCLIPro.bed.save("%s_%s_center.bed" % (out_fname_pref, read_grp), bed_middle, "%s center %s" % (read_grp, bam_basename), db_name="")

    ### generate read overlap maps
    fout_graph_pdf_fname = "%s_graphs.pdf" % out_fname_pref
    fout_graph_png_fname = "%s_graphs.png" % out_fname_pref
    fout_graph_tab_fname = "%s_graphs.tab" % out_fname_pref
    fout_report_fname = "%s_report.txt" % out_fname_pref
    fout_graph_tab = open(fout_graph_tab_fname, "wt")
    header = ["reads_group", "reference_group", "origin"]
    header = header + range(-group_comparisons_flank, group_comparisons_flank+1)
    fout_graph_tab.write("%s\n" % "\t".join([str(x) for x in header]))
    graph_color = '#0081c5'
    dy = 2
    dx = len(group_comparisons)
    plt.figure(figsize=(dx*1.9, dy*1.7))
    plt.clf()
    plt.subplots_adjust(left=0.15, right=0.9, bottom=0.12, top=0.9, wspace=0.55, hspace=0.7)
    si = 1
    for read_grp, ref_grp in group_comparisons:
        si_start = si
        for origin, grp_bed in [('start', grp_bed_start), ('center', grp_bed_middle)]:
            bed_cur = grp_bed.get(read_grp, {})
            bed_ref = grp_bed.get(ref_grp, {})

            pos_distrib = iCLIPro.bed.compare(bed_ref, bed_cur, group_comparisons_flank)
            if pos_distrib:
                xs, ys = zip(*sorted(pos_distrib.iteritems()))
            else:
                xs, ys = [], []

            ax = plt.subplot(dy, dx, si)
            plt.bar(xs, ys, width=0.7, ec=graph_color, color=graph_color)
            plt.xlim((-group_comparisons_flank, group_comparisons_flank))
            plt.xlabel("offset of %s to %s" % (read_grp, ref_grp))
            ax.get_xaxis().set_label_coords(0.5, -0.22)
            if (si - 1) % dx == 0:
                plt.ylabel("%s\nfrequency" % origin, multialignment='center', verticalalignment='center')
                ax.get_yaxis().set_label_coords(-0.4, 0.5)

            freqs = [0]*(2*group_comparisons_flank+1)
            for x, y in zip(xs, ys):
                assert -group_comparisons_flank<= x <= group_comparisons_flank
                i = group_comparisons_flank + x
                freqs[i] = y
            olst = [read_grp, ref_grp, origin] + freqs
            fout_graph_tab.write("%s\n" % "\t".join([str(x) for x in olst]))
            si += dx

        for _ in range(dx):
            plt.subplot(dy, dx, si_start)
            ylim = max(plt.ylim())
            plt.subplot(dy, dx, si_start+dx)
            ylim = max(ylim, max(plt.ylim()))
            ylim = (0.0, ylim)

            plt.plot([0, 0], ylim, c='k', lw=0.1, alpha=0.1)
            plt.ylim(ylim)
            yticks_vals, _ = plt.yticks()
            yticks_max = yticks_vals[-1]
            yticks = [0, yticks_max*1.0/4.0, yticks_max*2.0/4.0, yticks_max*3.0/4.0, yticks_max]
            plt.yticks(yticks, ["%0.0f" % y for y in yticks])

            xticks_vals, _ = plt.xticks()
            xticks_min = xticks_vals[0]
            xticks_max = xticks_vals[-1]
            xticks = [xticks_min, xticks_min*1.0/2.0]
            if xticks[-1] < 0:
                xticks.append(0)
            xticks = xticks + [xticks_max*1.0/2.0, xticks_max]
            plt.xticks(xticks, ["%0.0f" % x for x in xticks])

            plt.subplot(dy, dx, si_start)
            plt.plot([0, 0], ylim, c='k', lw=0.1, alpha=0.1)
            plt.ylim(ylim)
            plt.yticks(yticks, ["%0.0f" % y for y in yticks])
            plt.xticks(xticks, ["%0.0f" % x for x in xticks])

        si = si_start + 1
    fout_graph_tab.close()

    savefig_error_png = False
    #plt.savefig("%s_graphs.png" % (out_fname_pref))
    try:
        plt.savefig(fout_graph_png_fname)
    except:
        savefig_error_png = True
        print "ERROR: problems saving figure: '%s'" % fout_graph_png_fname
        print "       please check your matplotlib installation"
        if os.path.isfile(fout_graph_png_fname):
            os.remove(fout_graph_png_fname)

    savefig_error_pdf = False
    try:
        plt.savefig(fout_graph_pdf_fname)
    except:
        savefig_error_pdf = True
        print "ERROR: problems saving figure: '%s'" % fout_graph_pdf_fname
        print "       please check your matplotlib installation"
        if os.path.isfile(fout_graph_pdf_fname):
            os.remove(fout_graph_pdf_fname)

    ### main diagnostic
    # here we use the assumption that whatever origin (start or center) generates
    # less cross-linked positions, must be the correct one
    #
    # clustering of neighboring nucleotides is applied 
    # to minimize the variance when using read middle position
    #
    # only groups that do not containt full-length reads (max_read_len) are used
    better_interpretation = []
    for read_grp, valid_read_lens in sorted(g2lens.iteritems()):
        if iCLIPro.experimental:
            if max_read_len in valid_read_lens:
                # skip this one because it should not be used for this diagnostic
                continue

        start_bed = grp_bed_start.get(read_grp, {})
        middle_bed = grp_bed_middle.get(read_grp, {})

        for cl_size in neighbor_clustering:
            start_bed_clustered = iCLIPro.bed.cluster(start_bed, cl_size)
            middle_bed_clustered = iCLIPro.bed.cluster(middle_bed, cl_size)

            start_sites, start_cDNA = iCLIPro.bed.stats(start_bed_clustered)
            middle_sites, middle_cDNA = iCLIPro.bed.stats(middle_bed_clustered)

            if start_sites < middle_sites:
                pmode = 'start'
            elif start_sites > middle_sites:
                pmode = 'center'
            else:
                pmode = '='
            SM_ratio = float(start_sites)/max(1.0, middle_sites)
            better_interpretation.append((read_grp, cl_size, pmode, SM_ratio, start_sites, start_cDNA, middle_sites, middle_cDNA))

    ### generate report
    fout_report = open(fout_report_fname, "wt")

    fout_report.write("# Parameters\n")
    fout_report.write("#   in_bam_fname: %s\n" % in_bam_fname)
    fout_report.write("#   output_folder: %s\n" % output_folder)
    fout_report.write("#   bin_size: %s\n" % bin_size)
    fout_report.write("#   bin_min_reads: %s\n" % bin_min_reads)
    fout_report.write("#   neighbor_clustering: %s\n" % neighbor_clustering)
    fout_report.write("#   mapq_th: %s\n" % mapq_th)
    fout_report.write("#   split_flank: %s\n" % split_flank)
    fout_report.write("\n")

    fout_report.write("# Read filtering stats\n")
    fout_report.write("#   total number of records: %s\n" % all_recs)
    fout_report.write("#   not mapped: %s (%0.2f%% total)\n" % (notmapped_recs, 100.0*notmapped_recs/all_recs))
    fout_report.write("#   mapped: %s (%0.2f%% total)\n" % (mapped_recs, 100.0*mapped_recs/all_recs))
    fout_report.write("#   mapped but ignored because MAPQ < %s: %s (%0.2f%% total)\n" % (mapq_th, lowmapq_recs, 100.0*lowmapq_recs/all_recs))
    fout_report.write("#   mapped reads but ignored because gap in first/last %s nt: %s (%0.2f%% total)\n" % (split_flank, split_flank_ignored_recs, 100.0*split_flank_ignored_recs/all_recs))
    show_only = 30
    if len(all_bcs) > show_only:
        fout_report.write("#   distinct random barcodes (showing only %s out of all %s): %s\n" % (show_only, len(all_bcs), sorted(all_bcs)[:show_only]))
    else:
        fout_report.write("#   distinct random barcodes (%s): %s\n" % (len(all_bcs), sorted(all_bcs)))
    fout_report.write("\n") 

    fout_report.write("# Bin filtering stats\n")
    fout_report.write("#   total number of bins (of size %snt): %s\n" % (bin_size, tot_bins))
    fout_report.write("#   selected bins (with at least %s reads): %s (%0.2f%% total)\n" % (bin_min_reads, selected_bins, 100.0*selected_bins/tot_bins))
    fout_report.write("#   total reads: %s\n" % tot_reads)
    fout_report.write("#   selected reads (in selected bins): %s (%0.2f%% total)\n" % (selected_reads, 100.0*selected_reads/tot_reads))
    fout_report.write("#   selected reads length distribution and grouping:\n")
    fout_report.write("#\tread length\tnumber of reads\t% selected reads\tgroup(s)\n")
    notgrouped_reads = 0
    notgrouped_lengths = []
    for read_len, cn in sorted(selected_reads_len_distrib.iteritems()):
        if read_len not in readlen2groups:
            notgrouped_lengths.append(read_len)
            notgrouped_reads += cn
            grp_names = 'notused'
        else:
            grp_names = readlen2groups[read_len]
        fout_report.write("#\t%s\t%s\t%0.2f%%\t%s\n" % (read_len, cn, 100.0*cn/selected_reads, grp_names))
    if notgrouped_reads:
        fout_report.write("WARNING: %s reads (%0.2f%% selected) of lengths %s are not being used because not assigned to any group.\n" % (notgrouped_reads, 100.0*notgrouped_reads/selected_reads, notgrouped_lengths))
    fout_report.write("\n")

    fout_report.write("# Generated bedGraphs\n")
    fout_report.write("#\tgroup\torigin\ttotal cross-linked sites\ttotal cDNA\tfile name\n")
    for read_grp in sorted(g2lens.iterkeys()):
        bed_start = grp_bed_start.setdefault(read_grp, {})
        start_fname = "%s_%s_start.bed" % (out_fname_pref, read_grp)
        start_num_sites, start_tot_cDNA = iCLIPro.bed.stats(bed_start)

        bed_middle = grp_bed_middle.setdefault(read_grp, {})
        middle_fname = "%s_%s_center.bed" % (out_fname_pref, read_grp)
        middle_num_sites, middle_tot_cDNA = iCLIPro.bed.stats(bed_middle)

        fout_report.write("#\t%s\t%s\t%s\t%s\t%s\n" % (read_grp, 'start', start_num_sites, start_tot_cDNA, start_fname))
        fout_report.write("#\t%s\t%s\t%s\t%s\t%s\n" % (read_grp, 'center', middle_num_sites, middle_tot_cDNA, start_fname))
    fout_report.write("\n")

    fout_report.write("# Generated graphs for given comparisons\n")
    fout_report.write("#   graph data: %s\n" % (fout_graph_tab_fname))
    fout_report.write("#   graph: %s\n" % (fout_graph_pdf_fname))
    fout_report.write("#   graph: %s\n" % (fout_graph_png_fname))
    fout_report.write("\n")

    if iCLIPro.experimental:
        fout_report.write("# Which interpretation (start or center of read) generates\n")
        fout_report.write("# more consistent results (less cross-link sites)?\n")

    fout_report.write("#\tgroup\tneighbors clustered\tmode with least sites\tstart sites / center sites\tstart sites\tcenter sites\tstart cDNA\tcenter cDNA\n")
    pmode_freqs = {}
    for read_grp, cl_size, pmode, SM_ratio, start_sites, start_cDNA, middle_sites, middle_cDNA in better_interpretation:
        fout_report.write("#\t%s\t%s\t%s\t%0.3f\t%s\t%s\t%s\t%s\n" % (read_grp, cl_size, pmode, SM_ratio, start_sites, middle_sites, start_cDNA, middle_cDNA))
        pmode_freqs[pmode] = pmode_freqs.get(pmode, 0) + 1
    fout_report.write("#\n")
#    for cn, pmode in sorted([(cn, pmode) for pmode, cn in pmode_freqs.iteritems()], reverse=True):
#        fout_report.write("#   mode '%s' appears %s times\n" % (pmode, cn))
#    fout_report.write("\n")

    fout_report.write("END\n")
    fout_report.close() 

    for r in open(fout_report_fname, "rt"):
        print r.rstrip('\n\r')
    print

    if savefig_error_png:
        print "ERROR: problems saving figure: '%s'" % fout_graph_png_fname
        print "       please check your matplotlib installation"
    if savefig_error_pdf:
        print "ERROR: problems saving figure: '%s'" % fout_graph_pdf_fname
        print "       please check your matplotlib installation"
    print "Full report also available at: %s" % fout_report_fname

